"""Functions to query system's swap memory."""

import typing as t

from .available_features import psutil, SWAP


def query_swap(**_) -> t.Optional[t.Dict[str, t.Any]]:
    """Get information about swap."""
    if not SWAP:
        return None
    total_swap = psutil.swap_memory().total
    return {'total': total_swap}



"""Tests for hdd_info module."""

import unittest

from system_query.available_features import HDD
from system_query.hdd_info import query_hdds


@unittest.skipIf(not HDD, 'skipping HDD query')
class Tests(unittest.TestCase):

    def test_query_hdds(self):
        info = query_hdds()
        self.assertIsInstance(info, dict)
        self.assertGreaterEqual(len(info), 1)
        self.assertTrue(next(iter(info)).startswith('/dev'))



"""Tests for query and export utility function."""

import sys
import logging
import unittest

from system_query.query import query_and_export

_LOG = logging.getLogger(__name__)


class Tests(unittest.TestCase):

    def test_unsupported_scope(self):
        with self.assertRaises(NotImplementedError):
            query_and_export('host', 'raw', sys.stdout)
        with self.assertRaises(NotImplementedError):
            query_and_export('os', 'raw', sys.stdout)

    def test_unsupported_format(self):
        with self.assertRaises(NotImplementedError):
            query_and_export('all', 'xml', sys.stdout)
        with self.assertRaises(NotImplementedError):
            query_and_export('all', 'yaml', sys.stdout)

    def test_unsupported_target(self):
        with self.assertRaises(NotImplementedError):
            query_and_export('all', 'raw', 12345)
        with self.assertRaises(NotImplementedError):
            query_and_export('all', 'json', 12345)



"""Functions to query system's CPUs/APUs."""

import logging
import typing as t

from .available_features import cpuinfo, pint, psutil, CPU, CPU_CLOCK, CPU_CORES

_LOG = logging.getLogger(__name__)


def query_cpu_clock() -> t.Tuple[t.Optional[float], t.Optional[float], t.Optional[float]]:
    """Get current, minimum and maximum clock frequency of the CPU in the system."""
    if not CPU_CLOCK:
        return None, None, None
    try:
        cpu_clock = psutil.cpu_freq()
    except FileNotFoundError:
        return None, None, None
    if cpu_clock is None:
        return None, None, None
    return cpu_clock.current, cpu_clock.min, cpu_clock.max


def query_cpu_cores() -> t.Tuple[t.Optional[int], t.Optional[int]]:
    """Get number of logical and physical cores of the system's CPU."""
    if not CPU_CORES:
        return None, None
    return psutil.cpu_count(), psutil.cpu_count(logical=False)


def _get_cache_size(level: int, cpuinfo_data: dict) -> t.Optional[int]:
    """Get CPU cache size in bytes at a given level.

    If no units are provided, assume source data is in KiB.
    """
    raw_value = cpuinfo_data.get(
        f'l{level}_data_cache_size', cpuinfo_data.get(f'l{level}_cache_size', None))
    if raw_value is None:
        return None
    if isinstance(raw_value, int):
        return raw_value
    assert isinstance(raw_value, str), (type(raw_value), raw_value)
    # Sometimes cpuinfo prints e.g. `192 KiB (6 instances)` so we use
    # just the first two words
    raw_value = " ".join(raw_value.split()[:2])
    # KB, MB: "this practice frequently leads to confusion and is deprecated"
    # see https://en.wikipedia.org/wiki/JEDEC_memory_standards
    if raw_value.endswith('KB'):
        raw_value = raw_value[:-2] + 'KiB'
    elif raw_value.endswith('MB'):
        raw_value = raw_value[:-2] + 'MiB'
    ureg = pint.UnitRegistry()
    value = ureg(raw_value)
    if isinstance(value, int):
        return value * 1024
    assert isinstance(value, pint.Quantity), (type(value), value)
    if value.units == ureg('1').units:
        return value.magnitude * 1024
    _LOG.debug('L%i cache size parsed by pint: "%s" -> %s', level, raw_value, value)
    value = value.to('bytes')
    return int(value.magnitude)


def _get_cache_sizes(cpuinfo_data: dict) -> t.Mapping[int, t.Optional[int]]:
    return {lvl: _get_cache_size(lvl, cpuinfo_data) for lvl in range(1, 4)}


def query_cpu(**_) -> t.Mapping[str, t.Any]:
    """Get information about CPU present in the system."""
    if not CPU:
        return {}
    cpu = cpuinfo.get_cpu_info()
    clock_current, clock_min, clock_max = query_cpu_clock()
    logical_cores, physical_cores = query_cpu_cores()
    cache = _get_cache_sizes(cpu)
    return {
        'brand': cpu.get('brand', cpu.get('brand_raw', None)),
        'logical_cores': logical_cores,
        'physical_cores': physical_cores,
        'clock': clock_current,
        'clock_min': clock_min,
        'clock_max': clock_max,
        'cache': cache}



"""Tests for functionality that requires administrative access."""

import getpass
import unittest

from .test_all_info import IntegrationChecks


@unittest.skipUnless(getpass.getuser() == 'root', 'skipping tests that require sudo')
class Tests(IntegrationChecks):

    def test_query_all_sudo(self):
        self.query_all_and_check(True)



"""Query and export system data in one step."""

import json
import pathlib
import pprint
import sys
import typing as t

from .all_info import query_all
from .cpu_info import query_cpu
from .gpu_info import query_gpus
from .ram_info import query_ram
from .swap_info import query_swap

JSON_INDENT = 2

JSON_ENSURE_ASCII = False


def query_and_export(query_scope: str, export_format: str, export_target: t.Any, **kwargs):
    """Query the given scope of the system and export results in a given format to a given target.

    Currently implemented values are:

    - query_scope: all, cpu, gpu, ram, swap.
    - export_format: json, raw.
    - export_target: sys.stdout, sys.stderr, path.
    """
    info = query(query_scope, **kwargs)
    export(info, export_format, export_target)


QUERY_FUNCTIONS = {
    'all': query_all,
    'cpu': query_cpu,
    'gpu': query_gpus,
    'ram': query_ram,
    'swap': query_swap,
}


def query(query_scope: str, **kwargs) -> t.Any:
    """Wrap around selected system query functions."""
    try:
        query_function = QUERY_FUNCTIONS[query_scope]
    except KeyError as err:
        raise NotImplementedError(f'scope={query_scope}') from err
    return query_function(**kwargs)  # type: ignore


def export(info, export_format: str, export_target: t.Any):
    """Export information obtained by system query to a specified format."""
    if export_format == 'json':
        if export_target in (sys.stdout, sys.stderr):
            json_str = json.dumps(info, indent=JSON_INDENT, ensure_ascii=JSON_ENSURE_ASCII)
            print(json_str, file=export_target)
        elif isinstance(export_target, pathlib.Path):
            with open(str(export_target), 'w', encoding='utf-8') as json_file:
                json.dump(info, json_file, indent=JSON_INDENT, ensure_ascii=JSON_ENSURE_ASCII)
        else:
            raise NotImplementedError(f'format={export_format} target={export_target}')
    elif export_format == 'raw':
        if export_target in (sys.stdout, sys.stderr):
            pprint.pprint(info, stream=export_target)
        elif isinstance(export_target, pathlib.Path):
            with open(str(export_target), 'a', encoding='utf-8') as text_file:
                text_file.write(str(info))
        else:
            raise NotImplementedError(f'format={export_format} target={export_target}')
    else:
        raise NotImplementedError(f'format={export_format} target={export_target}')



"""Setup script for system_query package."""

import boilerplates.setup


class Package(boilerplates.setup.Package):
    """Package metadata."""

    name = 'system-query'
    description = 'Comprehensive and concise system information querying tool.'
    url = 'https://github.com/mbdevpl/system-query'
    author = 'Aleksandr Drozd, Emil Vatai, Mateusz Bysiek'
    maintainer = 'Mateusz Bysiek'
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: System :: Hardware',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Operating System',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
        'Typing :: Typed']
    keywords = ['system', 'software', 'hardware']
    extras_require = {
        'all': boilerplates.setup.parse_requirements('requirements_all.txt'),
        'cpu': boilerplates.setup.parse_requirements('requirements_cpu.txt'),
        'gpu': boilerplates.setup.parse_requirements('requirements_gpu.txt'),
        'hdd': boilerplates.setup.parse_requirements('requirements_hdd.txt'),
        'ram': boilerplates.setup.parse_requirements('requirements_memory.txt'),
        'swap': boilerplates.setup.parse_requirements('requirements_memory.txt')}


if __name__ == '__main__':
    Package.setup()



"""Command-line interface of system_query package."""

import argparse
import pathlib
import sys

from ._version import VERSION
from .query import QUERY_FUNCTIONS, query_and_export


def main(args=None, namespace=None):
    """Run the command-line interface.

    Execute query_and_export() function according to the arguments.
    """
    parser = argparse.ArgumentParser(
        prog='system_query', description='''Comprehensive and concise system information tool.
        Query a given hardware and/or software scope of your system and get results in human-
        and machine-readable formats.''',
        epilog='''Copyright 2017-2025 by the contributors, Apache License 2.0,
        https://github.com/mbdevpl/system-query''',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, allow_abbrev=True)
    parser.add_argument(
        '-s', '--scope', type=str, default='all', choices=list(QUERY_FUNCTIONS.keys()),
        help='''Scope of the query''')
    parser.add_argument(
        '-f', '--format', type=str, default='raw', choices=['raw', 'json'],
        help='''Format of the results of the query.''')
    parser.add_argument(
        '-t', '--target', type=str, default='stdout',
        help='''File path where to write the results of the query. Special values: "stdout"
        and "stderr" to write to stdout and stderr, respectively.''')
    parser.add_argument(
        '--version', action='version',
        version=f'{parser.prog} {VERSION},\nPython {sys.version}')

    args = parser.parse_args(args=args, namespace=namespace)
    target = {
        'stdout': sys.stdout,
        'stderr': sys.stderr
        }.get(args.target, pathlib.Path(args.target))
    query_and_export(query_scope=args.scope, export_format=args.format, export_target=target)



"""Probe the system for available features to establish what can be queried."""

import logging

_LOG = logging.getLogger(__name__)

try:
    import cpuinfo
except ImportError:
    cpuinfo = None
    _LOG.info("unable to import package cpuinfo", exc_info=True)
except:  # pylint: disable = bare-except  # noqa: E722
    # raise Exception("py-cpuinfo currently only works on X86 and some ARM CPUs.")
    cpuinfo = None  # pylint: disable = invalid-name
    _LOG.info("package cpuinfo doesn't work on this system", exc_info=True)

try:
    import pint
except ImportError:
    pint = None  # type: ignore
    _LOG.info("unable to import package pint", exc_info=True)

CPU = cpuinfo is not None and pint is not None

try:
    import psutil
except ImportError:
    psutil = None  # type: ignore
    _LOG.info("unable to import package psutil", exc_info=True)

CPU_CLOCK = psutil is not None and hasattr(psutil, 'cpu_freq')
CPU_CORES = psutil is not None

_GPU_FAILED = False
try:
    import pycuda
except ImportError:
    _GPU_FAILED = True
    _LOG.info("unable to import package pycuda", exc_info=True)
else:
    try:
        import pycuda.driver as cuda
    except ImportError:
        _GPU_FAILED = True
        _LOG.info("unable to import package pycuda.driver", exc_info=True)
    else:
        try:
            import pycuda.autoinit
        except ImportError:
            _GPU_FAILED = True
            _LOG.info("unable to import package pycuda.autoinit", exc_info=True)
        except pycuda._driver.Error:  # pylint: disable = protected-access
            _GPU_FAILED = True
            _LOG.info("unable to initialize cuda", exc_info=True)
        else:
            _LOG.debug('using CUDA version %s', '.'.join(str(_) for _ in cuda.get_version()))

if _GPU_FAILED:
    cuda = None  # pylint: disable = invalid-name

GPU = cuda is not None

try:
    import pyudev
except ImportError:
    pyudev = None
    _LOG.info("unable to import package pyudev", exc_info=True)
else:
    try:
        pyudev.Context()
    except ImportError:
        pyudev = None  # pylint: disable = invalid-name
        _LOG.info("udev library is not working", exc_info=True)

HDD = pyudev is not None

RAM_TOTAL = psutil is not None

SWAP = psutil is not None
